# filestore: Dmedia hashing protocol and file layout
# Copyright (C) 2012 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Generic API for the Dmedia Hashing Protocol, both current and future versions.

Dmedia is a very loosely coupled distributed object store, and in order for it
to achieve its goals, two key constraints were needed:

    1. Files in Dmedia are read-only (you can't modify an existing file, you can
       only insert an entire new file)

    2. Files can only be addressed by their content-hash

We feel the properties we get are well-worth these limitations, but assigning
global file IDs based on their content hash is not without a bit of peril.  For
some background on the challenges and limitations of content-based addressing,
see this excellent paper by Valerie Aurora:

    http://valerieaurora.org/monkey.html

Our take-away from the above paper is that:

    * We should assume the hashing protocol will have a limited useful
      lifetime, perhaps not longer than 10 years or so

    * We should today have in place exact plans for how a new version of the
      hashing protocol will be introduced and how multiple versions of the
      protocol can coexist side-by-side

In short, our migration strategy is to use the digest length as a way to select
the appropriate protocol version.  For example, the V1 hashing protocol has a
240-bit digest.  The V2 protocol, whenever it comes to be, will use a 280-bit
digest, or some other, larger multiple of 40 bits.

This allows files hashed under different versions of the protocol to be stored
together in the same `FileStore` layout.  At any given time there will be one or
more valid digest lengths, each one selecting a specific protocol version.  For
example, to verify every file in a `FileStore`, you would iterate through them
one by one, select the protocol version based on the ID length, and then check
if the hash you compute with that protocol matches the ID.

So that our reference implementation is well architected for the inevitable,
we're going to make all relevant code paths select the protocol version based
on the ID length, even while there is still only one protocol available.

We want to practice now, rather than later.  This will force us to work within
this constraint, and will mean we develop design patterns that are prepared for
the future.

Note that our original plan was to support the "interim protocol" (V0) alongside
V1, but the switch from RFC-3548 Base32 encoding to our Dbase32 encoding made
this impossible.  However, from V1 forward, all protocol version will be
supported indefinitely.

This module does more than just provide a V1 protocol implementation; it also
clearly defines our protocol framework, the design aspects that we expect to be
the same across all protocol versions.

Things that will (or can) change between protocol versions:

    * By definition, the digest size will change

    * The underlying hash function likely will change (to achieve better
      security, better performance, or hopefully both)

    * The leaf size might change (if it does, it will likely get bigger)

And things that we don't expect to change between protocol versions:

    * The digest size will always be a multiple of 40 bits

    * The root-hash and leaf-hash will use the same digest size (ie, if V2 uses
      a 320-bit root-hash, it will likewise use a 320-bit leaf-hash)

    * The leaf size will be a multiple of 2 MiB (likely a multiple of 8 MiB)

    * The root-hash will be cryptographically tied to the file-size, using the
      UTF-8 encoded decimal representation of the file-size as the key 

    * The leaf-hash will be cryptographically tied to the leaf-index, using
      the UTF-8 encoded decimal representation of the leaf-index as the key

    * All protocol versions will be non-configurable, meaning things like the
      leaf-size are strictly defined for each protocol version; for example, we
      wont support optionally using a 16 MiB leaf-size with the V1 protocol

    * Dbase32 will always be the "official" digest encoding, and will be the
      encoding use by standard `FileStore` layouts

Note that although the protocol framework defines the inputs that will be used
for the cryptographic tying, it does not define the exact mechanism, as the best
approach will vary depending upon the underlying hash function.  But the tying
must be done using a mechanism that would be suitable for MAC (for example, HMAC
with a SHA2 family hash).

We lucked out a bit with Skein as its elegant parameter system is a
fantastically good fit for our cryptographic tying.  Unfortunately, few hash
functions are defined with this sort of batteries-included pragmatism.
"""

from _skein import skein512


__all__ = ('VERSION0', 'VERSION1')

# Skein personalization strings used in the Dmedia Hashing Protocol:
PERS_LEAF = b'20110430 jderose@novacut.com dmedia/leaf'
PERS_ROOT = b'20110430 jderose@novacut.com dmedia/root'

# Additional Skein personalization strings used only in the old V0 protocol:
PERS_LEAF_INDEX = b'20110430 jderose@novacut.com dmedia/leaf-index'
PERS_FILE_SIZE = b'20110430 jderose@novacut.com dmedia/file-size'

# Provide very clear TypeError messages:
TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'

# Handy:
MiB = 1048576

# For now so we don't have any issues round-tripping values through JavaScript,
# we're setting a "soft" file-size limit at 2**53 bytes (the max integer that
# fully works in JavaScript).  But this is quite big, approximately 9.01 PB for
# a single file:
MAX_FILE_SIZE = 9007199254740992

# A real protocol would unlikely use a leaf-size this small, but a 2 MiB
# minimum means we can use a uint32_t for the leaf_index in C
# implementations and still reach the above MAX_FILE_SIZE:
MIN_LEAF_SIZE = 2097152


def make_key(integer):
    """
    Return *integer* as its UTF-8 encoded decimal representation.

    For example:

    >>> make_key(1776)
    b'1776'

    """
    if not isinstance(integer, int):
        raise TypeError(
            TYPE_ERROR.format('integer', int, type(integer), integer)
        )
    if integer < 0:
        raise ValueError('integer: must be >= 0; got {}'.format(integer))
    return str(integer).encode('utf-8')


class Protocol:
    """
    API for current and future versions of the Dmedia Hashing Protocol.

    This base class provides a standard API that abstracts away the details of
    a particular hashing protocol and allows multiple protocol versions to be
    supported simultaneously.

    The API consists of two methods:

        `Protocol.hash_leaf(leaf_index, leaf_data, challenge=b'')`

        `Protocol.hash_root(file_size, leaf_hashes)`

    And four attributes:

        `Protocol.leaf_size`

        `Protocol.digest_bits`

        `Protocol.digest_bytes`

        `Protocol.digest_b32len`

    To make it easy to experiment with different *leaf_size* and *digest_bits*
    (while still using the standard hashing functions), you can simply create
    a `SkeinProtocol` instance like this:

    >>> experimental = SkeinProtocol(16 * MiB, 360)

    The `digest_bytes` and `digest_b32len` will be derived from the provided
    *digest_bits*:

    >>> experimental.digest_bytes
    45
    >>> experimental.digest_b32len
    72

    This base class doesn't provide a working protocol, to do that you must
    create a subclass and implement these two methods:

        `Protocol._hash_leaf(key_leaf_index, leaf_data, challenge)`

        `Protocol._hash_root(key_file_size, leaf_hashes)` 
    """

    def __init__(self, leaf_size, digest_bits):
        if not isinstance(leaf_size, int):
            raise TypeError(
                TYPE_ERROR.format('leaf_size', int, type(leaf_size), leaf_size)
            )
        if leaf_size < MIN_LEAF_SIZE:
            raise ValueError('leaf_size: must be >= {}; got {}'.format(
                    MIN_LEAF_SIZE, leaf_size
                )
            )
        if leaf_size % MIN_LEAF_SIZE != 0:
            raise ValueError(
                'leaf_size: must be multiple of {}; got {}'.format(
                    MIN_LEAF_SIZE, leaf_size
                )
            )
        if not isinstance(digest_bits, int):
            raise TypeError(
                TYPE_ERROR.format(
                    'digest_bits', int, type(digest_bits), digest_bits
                )
            )
        if digest_bits < 200:
            raise ValueError(
                'digest_bits: must be >= 200; got {!r}'.format(digest_bits)
            )
        if digest_bits % 40 != 0:
            raise ValueError(
                'digest_bits: must be multiple of 40; got {!r}'.format(
                    digest_bits
                )
            )
        self.leaf_size = leaf_size
        self.max_leaf_count = MAX_FILE_SIZE // leaf_size
        self.digest_bits = digest_bits
        self.digest_bytes = digest_bits // 8
        self.digest_b32len = digest_bits // 5

    def hash_leaf(self, leaf_index, leaf_data, challenge=b''):
        if not isinstance(leaf_index, int):
            raise TypeError(
                TYPE_ERROR.format(
                    'leaf_index', int, type(leaf_index), leaf_index
                )
            )
        if not (0 <= leaf_index < self.max_leaf_count):
            raise ValueError(
                'Need 0 <= leaf_index < {!r}; got {!r}'.format(
                    self.max_leaf_count, leaf_index
                )
            )
        if not isinstance(leaf_data, bytes):
            raise TypeError(
                TYPE_ERROR.format(
                    'leaf_data', bytes, type(leaf_data), leaf_data
                )
            )
        if not (1 <= len(leaf_data) <= self.leaf_size):
            raise ValueError('Need 1 <= len(leaf_data) <= {}; got {}'.format(
                    self.leaf_size, len(leaf_data)
                )
            )
        return self._hash_leaf(make_key(leaf_index), leaf_data, challenge)

    def _hash_leaf(self, key_leaf_index, leaf_data, challenge):
        name = self.__class__.__name__
        raise NotImplementedError(
            '{}._hash_leaf(key_leaf_index, leaf_data, challenge)'.format(name)
        )

    def hash_root(self, file_size, leaf_hashes):
        if not isinstance(file_size, int):
            raise TypeError(
                TYPE_ERROR.format('file_size', int, type(file_size), file_size)
            )
        if not (1 <= file_size <= MAX_FILE_SIZE):
            raise ValueError(
                'Need 1 <= file_size <= {!r}; got {!r}'.format(
                    MAX_FILE_SIZE, file_size
                )
            )
        if not isinstance(leaf_hashes, bytes):
            raise TypeError(TYPE_ERROR.format(
                    'leaf_hashes', bytes, type(leaf_hashes), leaf_hashes
                )
            )
        if len(leaf_hashes) == 0:
            raise ValueError('leaf_hashes cannot be empty')
        if len(leaf_hashes) % self.digest_bytes != 0:
            raise ValueError(
                'len(leaf_hashes) is {}, not multiple of {}'.format(
                    len(leaf_hashes), self.digest_bytes
                )
            )
        count = len(leaf_hashes) // self.digest_bytes
        low = (count - 1) * self.leaf_size + 1
        high = count * self.leaf_size
        if not (low <= file_size <= high):
            raise ValueError(
                'Need {} <= file_size <= {}; got {}'.format(
                    low, high, file_size
                )
            )
        return self._hash_root(make_key(file_size), leaf_hashes)

    def _hash_root(self, key_file_size, leaf_hashes):
        name = self.__class__.__name__
        raise NotImplementedError(
            '{}._hash_root(key_file_size, leaf_hashes)'.format(name)
        )


class SkeinProtocol(Protocol):
    def _hash_leaf(self, key_leaf_index, leaf_data, challenge):
        return skein512(leaf_data,
            digest_bits=self.digest_bits,
            pers=PERS_LEAF,
            key=key_leaf_index,
            nonce=challenge,
        ).digest()

    def _hash_root(self, key_file_size, leaf_hashes):
        return skein512(leaf_hashes,
            digest_bits=self.digest_bits,
            pers=PERS_ROOT,
            key=key_file_size,
        ).digest()


class OldSkeinProtocol(Protocol):     
    def _hash_leaf_index(self, key_leaf_index):
        return skein512(key_leaf_index,
            digest_bits=self.digest_bits,
            pers=PERS_LEAF_INDEX,
        ).digest()

    def _hash_leaf(self, key_leaf_index, leaf_data, challenge):
        skein = skein512(self._hash_leaf_index(key_leaf_index),
            digest_bits=self.digest_bits,
            pers=PERS_LEAF,
            nonce=challenge,
        )
        skein.update(leaf_data)
        return skein.digest()

    def _hash_file_size(self, key_file_size):
        return skein512(key_file_size,
            digest_bits=self.digest_bits,
            pers=PERS_FILE_SIZE,
        ).digest()

    def _hash_root(self, key_file_size, leaf_hashes):
        skein = skein512(self._hash_file_size(key_file_size),
            digest_bits=self.digest_bits,
            pers=PERS_ROOT,
        )
        skein.update(leaf_hashes)
        return skein.digest()


# Create the Protocol instances:
VERSION0 = OldSkeinProtocol(8 * MiB, 240)
VERSION1 = SkeinProtocol(8 * MiB, 240)

